import json

from pytz import timezone

from fild.sdk import Field, Dictionary, dates


class DbTimestamp(Field):
    def generate_value(self):
        return dates.generate_time()

    def to_db(self):
        return self.value

    def to_format(self, fmt=dates.Pattern.DATE):
        return self.value.strftime(fmt)

    def to_timezone(self, tz):
        return self.value.astimezone(tz=timezone(tz))


class DBBaseJson(Dictionary):
    def with_values(self, values):
        if isinstance(values, str):
            values = json.loads(values)

        if values is not None:
            return super().with_values(values)

        return self



from sqlalchemy import create_engine, text
from sqlalchemy.orm import make_transient, sessionmaker


PLAY_EVOLUTIONS = 'play_evolutions'
FKEY_CHECK_OFF = 'SET FOREIGN_KEY_CHECKS=0;'
FKEY_CHECK_ON = 'SET FOREIGN_KEY_CHECKS=1;'
SYNC_COMMIT_OFF = 'SET SYNCHRONOUS_COMMIT TO OFF;'
TRUNC_ALL_TABLES = (
    "SELECT Concat('TRUNCATE TABLE ', table_schema, '.', TABLE_NAME, ';') "
    "FROM INFORMATION_SCHEMA.TABLES "
    "WHERE table_schema in ({0}) "
    "AND table_name NOT in ({1});"
)
TRUNC_ALL_TABLES_PG = (
    "SELECT "
    "Concat('TRUNCATE TABLE ', schemaname, '.', tablename, ' CASCADE;') "
    "FROM pg_tables "
    "WHERE schemaname in ({0}) "
    "AND tablename NOT in ({1});"
)


def to_dict(model_record, filter_none=True):
    d = {}

    for column in model_record.__table__.columns:
        column_name = column.name

        if column_name == 'global':
            column_name = 'is_global'

        if column_name == 'metadata':
            column_name = 'metadata_column'

        value = getattr(model_record, column_name)

        if filter_none and value is None:
            continue

        d[column_name] = value

    return d


class BaseClient:
    connection = None

    def cascade_delete(self, model):
        sql = f'TRUNCATE {model.__table__.__tablename__} CASCADE;'
        self.connection.execute(sql)
        self.connection.commit()
        self.connection.close_all()

    def update(self, model, new_values, *criteria, **kwargs):
        """
        Note: new_values - a dictionary where keys are column names,
         values - corresponding values to set.
        """
        query = self.connection.query(model.__table__)

        if criteria:
            records = query.filter(*criteria)
        else:
            records = query.filter_by(**kwargs)

        records.update(new_values, synchronize_session='fetch')
        self.connection.commit()
        self.connection.close_all()

    def delete(self, model, *criteria, **kwargs):
        """
        :param criteria: Conditional criteria to delete records, e.g.:
          MyClass.name == 'some name'
          MyClass.id > 5,
          MyClass.field.in_([1, 2, 3])
        :param kwargs: Key-value conditions, e.g.:
          name='some name'
          id=5
        """
        query = self.connection.query(model.__table__)

        if criteria:
            query = query.filter(*criteria)
        else:
            query = query.filter_by(**kwargs)

        query.delete(synchronize_session=False)
        self.connection.commit()
        self.connection.close_all()

    def insert(self, record):
        model = record
        record = model.to_table_record()

        self.connection.add(record)
        self.connection.commit()
        # refresh() gets actual record state after commit
        # (needed to make_transient)
        # make_transient unbinds model from slqalchemy session
        self.connection.refresh(record)
        make_transient(record)
        self.connection.close_all()

        return model.__class__(is_custom=model.is_custom).with_values(
            to_dict(record)
        )

    def pre_insert(self, record):
        self.connection.add(record)
        self.connection.flush()

    def commit_and_close(self):
        self.connection.commit()
        self.connection.close_all()

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        raise NotImplementedError


class ConnectionClient(BaseClient):
    _db = None

    def execute(self, sql_script, *args):

        if args:
            result = self.connection.execute(sql_script, args)
        else:
            result = self.connection.execute(text(sql_script))

        if result.returns_rows:
            return result.fetchall()

        return None

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        raise NotImplementedError


class PostgresqlDBClient(ConnectionClient):
    def __init__(self, host='0.0.0.0', port=5432, user='user', password=None,
                 db='db', sync_commit=False, **kwargs):
        self._db = 'public'
        self.connection_url = (
            f'postgresql://{user}:{password}@{host}:{port}/{db}'
        )
        self.connection = None
        self._kwargs = kwargs
        self._sync_commit = sync_commit

    def connect(self):
        if self.connection is None:
            engine = create_engine(
                self.connection_url
            )
            session = sessionmaker(bind=engine)
            self.connection = session()

            if not self._sync_commit:
                self.execute(SYNC_COMMIT_OFF)
                self.connection.commit()
                self.connection.close()

        return self

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        exclude_tables = exclude_tables or []
        exclude_str = "', '".join(exclude_tables)
        exclude_tables = f"'{exclude_str}'"
        schemas = schemas or [self._db]

        if self._db not in schemas:
            schemas.append(self._db)

        schemas_str = "', '".join(schemas)
        db_schemas = f"'{schemas_str}'"

        sql = TRUNC_ALL_TABLES_PG.format(db_schemas, exclude_tables)
        trunc_stmts = ''

        for result in self.execute(sql):
            trunc_stmts += result[0]

        self.execute(trunc_stmts)


class MysqlDBClient(ConnectionClient):
    def __init__(self, host='127.0.0.1', port=3306, user='root',
                 password='mypass', db_name='testengine'):
        self._db = db_name
        self.connection_url = (
            f'mysql://{user}:{password}@{host}:{port}/'
            f'{db_name}?charset=utf8mb4'
        )
        self.connection = None

    def connect(self):
        if self.connection is None:
            engine = create_engine(self.connection_url)
            session = sessionmaker(bind=engine)
            self.connection = session()

        return self

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        # Test shouldn't truncate internal evolutions table
        exclude_tables = exclude_tables or []
        exclude_tables.append(PLAY_EVOLUTIONS)
        exclude_str = "', '".join(exclude_tables)
        exclude_tables = f"'{exclude_str}'"
        schemas = schemas or [self._db]

        if self._db not in schemas:
            schemas.append(self._db)

        schemas_str = "', '".join(schemas)
        db_schemas = f"'{schemas_str}'"

        sql = TRUNC_ALL_TABLES.format(db_schemas, exclude_tables)
        trunc_stmts = ''

        for result in self.connection.execute(sql):
            trunc_stmts += result[0]

        sql = ''.join([FKEY_CHECK_OFF, trunc_stmts, FKEY_CHECK_ON])
        self.execute(sql)


class SqliteDBClient(BaseClient):
    def __init__(self, file_path):
        self._db_file_path = file_path

    def connect(self):
        import sqlite3  # pylint: disable=import-outside-toplevel

        if self.connection is None:
            self.connection = sqlite3.connect(self._db_file_path)

        return self

    def close_connection(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def read_data(self, sql_script, *args):
        return self.connection.execute(sql_script, args).fetchall()

    def write_data(self, sql_script, *args):
        self.connection.execute(sql_script, args)
        self.connection.commit()

    def __enter__(self):
        self.connect()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_connection()

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        raise NotImplementedError


class CassandraDBClient(BaseClient):
    def __init__(self, hosts=None, keyspace='cqlengine', protocol_version=3):
        self.hosts = hosts or ['localhost']
        self.keyspace = keyspace
        self.protocol_version = protocol_version

    def connect(self):
        if self.connection is None:
            from cassandra.cqlengine import connection # pylint: disable=import-outside-toplevel
            self.connection = connection
            connection.setup(
                self.hosts,
                self.keyspace,
                protocol_version=self.protocol_version,
                lazy_connect=True
            )

        return self

    def insert(self, record):
        self.connect()
        record.__table__.create(**record.to_db())

        return record

    def cascade_delete(self, model):
        raise NotImplementedError

    def update(self, model, new_values, *criteria, **kwargs):
        raise NotImplementedError

    def delete(self, model, *criteria, **kwargs):
        raise NotImplementedError


class DbClient(BaseClient):
    __clients = {}

    def __new__(cls, client_name, client, *args, **kwargs) -> BaseClient:  # pylint: disable=unused-argument
        if client_name not in DbClient.__clients:
            DbClient.__clients[client_name] = client.connect()

        return DbClient.__clients[client_name]

    def trunc_all_tables(self, schemas=None, exclude_tables=None):
        raise NotImplementedError



from waiting import wait

from fild_compare import compare
from fild_db.client import DbClient, to_dict
from fild_db.types.model import DbModel


DEFAULT_DB_TIMEOUT = 3


class Database:
    _no_db_mode = False

    def __init__(self, client_name, client):
        self.db = DbClient(client_name=client_name, client=client)

    def enable_no_db_mode(self):
        self._no_db_mode = True

    def reset_mode(self):
        self._no_db_mode = False

    def _get_records(self, model, *criteria, **kwargs):
        order_by = kwargs.pop('order_by', None)
        query = self.db.connection.query(model)

        if criteria:
            data = query.filter(*criteria).filter_by(**kwargs).order_by(
                order_by
            ).all()
        else:
            data = query.filter_by(**kwargs).order_by(order_by).all()

        self.db.connection.close()
        return data

    def get_record(self, model, *criteria, **kwargs):
        return self.get_records(model, *criteria, **kwargs)[0]

    def get_records_nowait(self, model, *criteria, **kwargs):
        return [
            model(is_custom=True).with_values(to_dict(rec))
            for rec in self._get_records(model.__table__, *criteria, **kwargs)
        ]

    def get_records(self, model, *criteria, **kwargs):
        sleep_seconds = kwargs.pop('sleep_seconds', 0)
        timeout_seconds = kwargs.pop('timeout_seconds', None)

        def filter_records():
            return [
                model(is_custom=True).with_values(to_dict(rec))
                for rec in self._get_records(
                    model.__table__, *criteria, **kwargs
                )
            ]

        return wait(
            filter_records,
            waiting_for=f'records from {model.get_table_name()} by: {kwargs}',
            timeout_seconds=timeout_seconds or DEFAULT_DB_TIMEOUT,
            sleep_seconds=sleep_seconds
        )

    def insert(self, record):
        if self._no_db_mode:
            return None

        return self.db.insert(record)

    def insert_records(self, records):
        if self._no_db_mode:
            return

        for record in records:
            record = record.to_table_record()
            self.db.pre_insert(record)

        self.db.commit_and_close()

    def delete(self, model, *criteria, **kwargs):
        """
        :param criteria: Conditional criteria to delete records, e.g.:
          MyClass.name == 'some name'
          MyClass.id > 5,
          MyClass.field.in_([1, 2, 3])
        :param kwargs: Key-value conditions, e.g.:
          name='some name'
          id=5
        """
        self.db.delete(model, *criteria, **kwargs)

    def update(self, model, new_values, *criteria, **kwargs):
        """
        Note: new_values - a dictionary where keys are column names,
         values - corresponding values to set.
        """
        self.db.update(model, new_values, *criteria, **kwargs)

    def cascade_delete(self, model):
        self.db.cascade_delete(model)

    def verify_no_record(self, model, *criteria, **kwargs):
        data = self._get_records(model.__table__, *criteria, **kwargs)
        assert not data, (
            f'Unexpected {model.get_table_name()} record by: {kwargs}'
        )

    def verify_no_record_with_wait(self, model, *criteria, **kwargs):
        wait(
            lambda: not self._get_records(model.__table__, *criteria, **kwargs),
            waiting_for=f'no {model.get_table_name()} records by: {kwargs}',
            timeout_seconds=DEFAULT_DB_TIMEOUT,
            sleep_seconds=0
        )

    @staticmethod
    def verify_record(actual: DbModel, expected: DbModel, rules=None):
        compare(
            actual=actual.value,
            expected=expected.value,
            rules=rules
        )

    @staticmethod
    def verify_records(actual: [DbModel], expected: [DbModel], rules=None):
        target_name = ''

        if actual:
            target_name = actual[0].get_table_name()
        elif expected:
            target_name = expected[0].get_table_name()

        actual_data = [item.value for item in actual]
        expected_data = [item.value for item in expected]
        compare(
            actual=actual_data,
            expected=expected_data,
            target_name=f'{target_name} records',
            rules=rules
        )

    def trunc_all_tables(self, schemas=None, exclude=None):
        self.db.trunc_all_tables(schemas=schemas, exclude_tables=exclude) # pylint: disable=no-member



from fild.sdk import Dictionary, Field


class DbModel(Dictionary):
    __table__ = None

    def __init__(self, is_custom=False, is_full=False):
        self.is_custom = is_custom
        self.save_kwargs(locals())
        super().__init__(is_full=is_full)

    @classmethod
    def get_table_name(cls):
        return cls.__table__.__tablename__

    def _is_column_required(self, column_name):
        if column_name == 'global':
            column_name = 'is_global'

        if column_name == 'metadata':
            column_name = 'metadata_column'

        column = getattr(self.__table__, column_name)
        return not column.nullable or column.default

    def _generate(self, is_full=False, with_data=True, required=True,
                  use_default=None):
        self._generated = with_data

        for field_name in self._get_field_names():

            value = getattr(self, field_name)
            _with_data = with_data

            if _with_data:
                if self.is_custom:
                    _with_data = self._is_column_required(value.name)
                else:
                    _with_data = not (isinstance(value, Field) and
                                      not value.required and not is_full)

            setattr(self, field_name, value(
                is_full=is_full, with_data=_with_data
            ))

        if with_data:
            self.generate_custom()

    def to_db(self):
        _value = {}

        for field_name in self._get_field_names():
            field_value = getattr(self, field_name)

            if field_value.generated:
                name = field_value.name

                if name == 'global':
                    name = 'is_global'

                if name == 'metadata':
                    name = 'metadata_column'

                if hasattr(field_value, 'to_db'):
                    _value[name] = field_value.to_db()
                else:
                    _value[name] = field_value.value

        return _value

    def to_table_record(self):
        return self.__table__(**self.to_db())  # pylint: disable=not-callable


class CassandraDbModel(DbModel):
    @classmethod
    def get_table_name(cls):
        return cls.__table__.__table_name__

    def to_table_record(self):
        raise NotImplementedError
